module.exports = SelectionView;

// #### <a name="view"></a> `SelectionView`
//
// The `SelectionView` is used internally to represent the selection view, that is the
// visual selection of either one or more cells.
function SelectionView (table, selection) {
    var self = this,
        html = document.getElementsByTagName('html')[0];

    self.element = document.createElement('div');
    self.element.style.position = 'absolute';
    self.element.style.display = 'none';
    self.element.tabIndex = -1;

    self.inputElement = document.createElement('input');
    self.inputElement.style.position = 'absolute';
    self.inputElement.style.display = 'none';

    self.copyPasteElement = document.createElement('textarea');
    self.copyPasteElement.style.position = 'absolute';
    self.copyPasteElement.style.opacity = '0.0';
    self.copyPasteElement.style.display = 'none';

    table.appendChild(self.element);
    table.appendChild(self.inputElement);
    table.appendChild(self.copyPasteElement);

    self.destroy = function () {
        self.element.removeEventListener('mousedown', self.onMouseDown);
        self.element.removeEventListener('dblclick', self.onDblClick);
        self.element.removeEventListener('keypress', self.onKeyPress);
        self.element.removeEventListener('keydown', self.onKeyDown);

        self.inputElement.removeEventListener('keydown', self.onInputKeydown);
        self.inputElement.removeEventListener('blur', self.onInputBlur);

        $(html).unbind('mouseup', self.onMouseUp);

        table.removeChild(self.element);
        table.removeChild(self.inputElement);
        table.removeChild(self.copyPasteElement);
    };
    self.show = function () {
        self.element.style.display = 'block';
        self.element.focus();
        if (self.element.scrollIntoViewIfNeeded) {
            self.element.scrollIntoViewIfNeeded();
        }
        else {
            self.element.scrollIntoView();
        }
    };
    self.hide = function () {
        self.element.style.display = 'none';
    };
    self.focus = function () {
        self.element.focus();
    };
    self.update = function (start, end) {
        var top = Math.min(start.offsetTop, end.offsetTop),
            left = Math.min(start.offsetLeft, end.offsetLeft),
            bottom = Math.max(start.offsetTop + start.offsetHeight,
                            end.offsetTop + end.offsetHeight),
            right = Math.max(start.offsetLeft + start.offsetWidth,
                            end.offsetLeft + end.offsetWidth);

        self.element.style.top = top + 1 + 'px';
        self.element.style.left = left + 1 + 'px';
        self.element.style.height = bottom - top - 1 + 'px';
        self.element.style.width = right - left - 1 + 'px';
        self.element.style.backgroundColor = 'rgba(245, 142, 00, 0.15)';

        self.show();
    };
    self.beginDrag = function () {
        self.canDrag = true;
        ko.utils.registerEventHandler(self.element, 'mousemove', self.doBeginDrag);
    };
    self.doBeginDrag = function () {
        self.element.removeEventListener('mousemove', self.doBeginDrag);

        if (!self.canDrag) {
            return;
        }

        self.isDragging = true;
        self.element.style.pointerEvents = 'none';
    };
    self.endDrag = function () {
        self.element.removeEventListener('mousemove', self.doBeginDrag);
        self.isDragging = false;
        self.canDrag = false;
        self.element.style.pointerEvents = 'inherit';
    };

    self.onMouseUp = function (event) {
        self.endDrag();
    };
    self.onMouseDown = function (event) {
        if (event.button !== 0) {
            return;
        }

        self.hide();

        var cell = event.view.document.elementFromPoint(event.clientX, event.clientY);
        selection.onCellMouseDown(cell, event.shiftKey);

        event.preventDefault();
    };
    self.onDblClick = function (event) {
        selection.startEditing();
    };
    self.onKeyPress = function (event) {
        selection.startEditing();
    };
    self.onKeyDown = function (event) {
        if (event.keyCode === 13) {
            selection.onReturn(event);
        } else if ([37, 38, 39, 40].indexOf(event.keyCode) !== -1) {
            selection.onArrows(event);
        } else if (event.keyCode === 86 && event.ctrlKey) {
            self.copyPasteElement.value = '';
            self.copyPasteElement.style.display = 'block';
            self.copyPasteElement.focus();
            setTimeout(function () {
                selection.onPaste(self.copyPasteElement.value);
                self.copyPasteElement.style.display = 'none';
                self.focus();
            }, 0);
        } else if (event.keyCode === 67 && event.ctrlKey) {
            self.copyPasteElement.value = selection.onCopy();
            self.copyPasteElement.style.display = 'block';
            self.copyPasteElement.focus();
            document.execCommand('selectAll', false, null);
            setTimeout(function () {
                self.copyPasteElement.style.display = 'none';
                self.focus();
            }, 0);
        } else if (event.keyCode === 9) {
            selection.onTab(event);
        }
    };
    self.onInputKeydown = function (event) {
        var cell = selection.range.start;

        if (event.keyCode === 13) { // Return
            var value = selection.endEditingCell(cell);

            if (event.ctrlKey) {
                ko.utils.arrayForEach(selection.range.getCells(), function (cellInSelection) {
                if (cellInSelection !== cell) {
                    selection.updateCellValue(cellInSelection, value);
                }
                });
          }

            selection.onReturn(event, event.ctrlKey);
            self.focus();
            event.preventDefault();
        }
        else if (event.keyCode === 27) { // Escape
            selection.cancelEditingCell(cell);
            self.focus();
        }
    };
    self.onInputBlur = function (event) {
        if (!selection.isEditingCell()) {
            return;
        }
        selection.endEditingCell(selection.range.start);
    };

    ko.utils.registerEventHandler(self.element, "mousedown", self.onMouseDown);
    ko.utils.registerEventHandler(self.element, "dblclick", self.onDblClick);
    ko.utils.registerEventHandler(self.element, "keypress", self.onKeyPress);
    ko.utils.registerEventHandler(self.element, "keydown", self.onKeyDown);

    ko.utils.registerEventHandler(self.inputElement, "keydown", self.onInputKeydown);
    ko.utils.registerEventHandler(self.inputElement, "blur", self.onInputBlur);

    ko.utils.registerEventHandler(html, "mouseup", self.onMouseUp);
}
